/*Functions for managing types needed to construct SNMP messages*/
#include <sys/types.h>
#include <netinet/in.h>

#include "Message.h"
#include "error.h"
#include "wrap.h"
#include "mes_func.h"
const int arcsize= sizeof(long);


const struct PDU_handler {
	PDUs_PR		pdu_type;
	int	(*pdu_t_cr_func) __P((PDU_Choice_t *,mesg_data_p));
	int	(*pdu_t_putrefy_func) __P((PDU_Choice_t *,mesg_data_p));
	int	(*pdu_t_free_func) __P((PDU_Choice_t *));
} pdu_handlers[] = {
	{ PDUs_PR_NOTHING,		NULL,			NULL,				NULL},
	{ PDUs_PR_get_request,  	get_request_cr,		get_request_putrefy,		get_request_free},	
	{ PDUs_PR_get_next_request,	get_next_cr,		get_next_putrefy,		get_next_free},
	{ PDUs_PR_get_bulk_request,	NULL,			NULL,				NULL},
	{ PDUs_PR_response,		response_request_cr,	response_request_putrefy,	response_request_free},
	{ PDUs_PR_set_request,		set_request_cr,		set_request_putrefy,		set_request_free},
	{ PDUs_PR_inform_request,	NULL,			NULL,				NULL},
	{ PDUs_PR_snmpV2_trap,		NULL,			NULL,				NULL},
	{ PDUs_PR_report,		NULL,			NULL,				NULL},
};

const struct VarBindValue_handler {
	VarBindValue_PR vbv_type;
	int	(*vbv_cr_func) __P((VBV_Choice_t *,vb_data_p ));
	int	(*vbv_putrefy_func) __P((VBV_Choice_t *,vb_data_p));
	int	(*vbv_cpv_func) __P((char *,vb_data_p));
	int	(*vbv_free_func) __P((VBV_Choice_t *));
} vbv_handlers[] = {
	{ VarBindValue_PR_NOTHING,		NULL,			NULL,			NULL,		NULL	},	
	{ VarBindValue_PR_value,		obj_syn_cr,		obj_syn_putrefy,	obj_syn_cpv,	obj_syn_free	},
	{ VarBindValue_PR_unSpecified,		vbv_unspecified_cr,	NULL,			nothing_cpv,	vbv_unspecified_free	},
	{ VarBindValue_PR_noSuchObject,		NULL,			NULL,			NULL,		NULL},
	{ VarBindValue_PR_noSuchInstance,	NULL,			NULL,			NULL,		NULL},
	{ VarBindValue_PR_endOfMibView,		NULL,			NULL,			NULL,		NULL},
};

const struct ObjectSyntax_handler {
	ObjectSyntax_PR		os_type;
	int	(*os_cr_func) __P((OS_Choice_t *,vb_data_p ));
	int	(*os_putrefy_func) __P((OS_Choice_t *,vb_data_p ));
	int	(*os_cpv_func) __P((char *,vb_data_p ));
	int	(*os_free_func) __P((OS_Choice_t *));
} os_handlers[] = {
	{ ObjectSyntax_PR_NOTHING,		NULL,		NULL,			NULL,		NULL},
	{ ObjectSyntax_PR_simple,		sim_syn_cr,	sim_syn_putrefy,	sim_syn_cpv,	sim_syn_free},
	{ ObjectSyntax_PR_application_wide,	NULL,		NULL,			NULL,		NULL},
};

const struct SimpleSyntax_handler {
	SimpleSyntax_PR		ss_type;
	int	(*ss_cr_func) __P((SS_Choice_t *,OCTET_STRING_t *));
	int	(*ss_putrefy_func) __P((SS_Choice_t *,OCTET_STRING_t *));
	int	(*ss_cpv_func) __P((char *,vb_data_p));
	int	(*ss_free_func) __P((SS_Choice_t *));
} ss_handlers[] = {
	{ SimpleSyntax_PR_NOTHING,		NULL,		NULL,		NULL,		NULL},
	{ SimpleSyntax_PR_integer_value,	int_cr, 	int_putrefy,	int_cpv,	int_free},
	{ SimpleSyntax_PR_string_value,		o_str_cr,	o_str_putrefy,	o_str_cpv,	o_str_free},
	{ SimpleSyntax_PR_objectID_value,	oid_cr,		oid_putrefy,	oid_cpv,	oid_free},
};

void
free_mesg_data(mesg_data_p p) {
    int i;

    for (i=0;i<p->vb_n;i++){
	if (p->data+i) {
	    free_vb_data(p->data+i);
	}
    }
    if (p->data){
	free(p->data);
    }
    if (p->community.buf)
	free(p->community.buf);
    free(p);
}

void
free_vb_data(vb_data_p p){
    if (p->oid.buf){
	free(p->oid.buf);
    }
    if (p->value.buf) {
	free(p->value.buf);
    }
}

int
free_message(Message_t * mesg) {
    int res;

    asn_DEF_Message.free_struct(&asn_DEF_Message,mesg,0);
    return res;
}


int
fill_mesg(Message_t * mesg, mesg_data_p p) {

    mesg->version = p->version;
    if ((OCTET_STRING_fromBuf(&(mesg->community), p->community.buf, p->community.size))!=0) {
	err_msg("couldn't make community");
	return -1;
    }
    return fill_mesg_data(&(mesg->data),p);
}

int
putrefy_mesg(Message_t *mesg, mesg_data_p p) {

    p->version=mesg->version;
    if ((OCTET_STRING_fromBuf(&(p->community), mesg->community.buf, mesg->community.size))!=0) {
	err_msg("couldn't make community");
	return -1;
    }
    return putrefy_mesg_data(&(mesg->data),p);
}

int
compare_mesg(mesg_data_p p,mesg_data_p q){

    if (q->ip.s_addr!=p->ip.s_addr) {
	err_msg("received message from another device");
	return -1;
    }
    if (q->r_id != p->r_id) {
	err_msg("received message with different request_id");
	return -1;
    }
    if (q->vb_n != p->vb_n) {
	err_msg("received message with different variable binding count");
	return -1;
    }
    if ((q->pdu_pr==PDUs_PR_get_next_request) || (p->pdu_pr==PDUs_PR_get_next_request)){
	return 0;
    }
    else {
	return compare_vb_data(p->data, q->data, p->vb_n);
    }
}

int
compare_vb_data (vb_data_p p, vb_data_p q, int number){
    int i;
    
    for (i=0;i<number; i++) {
	if (compare_oid(p->oid,q->oid)!=0){
	    err_msg("received message with different oids in %i binding",i+1);
	    return -1;
	}
    }
    return 0;
}

int
compare_oid (OBJECT_IDENTIFIER_t p, OBJECT_IDENTIFIER_t q){
    int i;

    if (q.size!=p.size) {
	err_msg("received message with different oid sizes");
	return -1;
    }
    for (i=0;i<q.size;i++) {
	if (q.buf[i]!=p.buf[i]) {
	    err_msg("received message with different oid bytes in position %i",i);
	    return -1;
	}
    }
    return 0;
}


int
fill_mesg_data(PDUs_t *pdu,mesg_data_p p) {

    pdu->present=p->pdu_pr;
    if (pdu_handlers[pdu->present].pdu_t_cr_func != NULL){
	
	return (*(pdu_handlers[pdu->present]).pdu_t_cr_func)(&(pdu->choice),p);
    }
    else {
	err_msg("PDU type not implemented yet\n");
	return -1;
    }
}

int
free_message_data(PDUs_t *pdu) {
    if (pdu_handlers[pdu->present].pdu_t_free_func != NULL){
	
	return (*(pdu_handlers[pdu->present]).pdu_t_free_func)(&(pdu->choice));
    }
    else {
	err_msg("PDU type not implemented yet\n");
	return -1;
    }
}


int
putrefy_mesg_data(PDUs_t *pdu,mesg_data_p p) {

    p->pdu_pr=pdu->present;
    if (pdu_handlers[pdu->present].pdu_t_putrefy_func != NULL){
	return (*(pdu_handlers[pdu->present]).pdu_t_putrefy_func)(&(pdu->choice),p);
    }
    else {
	err_msg("PDU type not implemented yet\n");
	return -1;
    }
}

int
pdu_cr(PDU_t *pointer,mesg_data_p p) {
    int res=-1;

    pointer->request_id=p->r_id;
    pointer->error_status=0;
    pointer->error_index=0;
    if ((res=vbl_cr(&(pointer->variable_bindings),p))== -1)
	err_msg("Couldn't create VB List");
    return res;
}

int
pdu_free(PDU_t *pointer) {
    int res;

    printf("Freeing pdu");
    if ((res=vbl_free(&(pointer->variable_bindings)))==-1)
	err_msg("Couldn't free VB List");
    printf("pdu freed");
    return res;
}

int
pdu_putrefy(PDU_t *pointer,mesg_data_p p) {
    int res=-1;

    p->r_id=pointer->request_id;
    if(pointer->error_status){
	err_msg("Received erroroneous message at %i",pointer->error_index);	
    }
    p->error_status=pointer->error_status;
    p->error_index=pointer->error_index;
    if(p->error_status){
	err_msg("Received erroroneous message at %i",p->error_index);	
    }
    if ((res=vbl_putrefy(&(pointer->variable_bindings),p))== -1)
	err_msg("Couldn't reconstruct VB List from message");
    return res;
}


int
set_request_cr(PDU_Choice_t *upointer,mesg_data_p p) {
    PDU_t *pointer=&(upointer->set_request);
    int res=-1;

    if ((res=pdu_cr(pointer,p))==-1)
	err_msg("Couldn't create set-request");
    return res;
}

int
set_request_free(PDU_Choice_t *upointer){
    PDU_t *pointer=&(upointer->set_request);
    int res;
    if ((res=pdu_free(pointer))==-1)
	err_msg("Couldn't create set-request");
    return res;
}

int
get_request_cr(PDU_Choice_t *upointer,mesg_data_p p) {
    PDU_t *pointer=&(upointer->get_request);
    int res=-1;

    if ((res=pdu_cr(pointer,p))==-1)
	err_msg("Couldn't create get-request");
    return res;
}

int
get_request_free(PDU_Choice_t *upointer){
    PDU_t *pointer=&(upointer->get_request);
    int res=-1;

    if ((res=pdu_free(pointer))==-1)
	err_msg("Couldn't create set-request");
    return res;
}

int
get_next_cr(PDU_Choice_t *upointer,mesg_data_p p) {
    PDU_t *pointer=&(upointer->get_next_request);
    int res=-1;

    if ((res=pdu_cr(pointer,p))==-1)
	err_msg("Couldn't create get-next-request");
    return res;
}

int
get_next_free(PDU_Choice_t *upointer){
    PDU_t *pointer=&(upointer->get_next_request);
    int res=-1;

    printf("Freeing get Next");
    if ((res=pdu_free(pointer))==-1)
	err_msg("Couldn't create set-request");
    printf("Freed get-Next");
    return res;
}

int
set_request_putrefy(PDU_Choice_t *upointer,mesg_data_p p) {
    PDU_t *pointer=&(upointer->set_request);
    int res=-1;

    if ((res=(pdu_putrefy(pointer,p)))==-1)
	err_msg("Couldn't reconstruct set-request");
    return res;
}

int
response_request_cr(PDU_Choice_t *upointer,mesg_data_p p) {
    PDU_t *pointer=&(upointer->response);
    int res=-1;

    if ((res=pdu_cr(pointer,p))==-1)
	err_msg("Couldn't create responce-request");
    return res;
}

int
response_request_free(PDU_Choice_t *upointer){
    PDU_t *pointer=&(upointer->response);
    int res=-1;

    if ((res=pdu_free(pointer))==-1)
	err_msg("Couldn't create set-request");
    return res;
}


int
response_request_putrefy(PDU_Choice_t *upointer,mesg_data_p p) {
    PDU_t *pointer=&(upointer->response);
    int res=-1;

    if ((res=(pdu_putrefy(pointer,p)))==-1)
	err_msg("Couldn't reconstruct responce-request");
    return res;
}

int
get_request_putrefy(PDU_Choice_t *upointer,mesg_data_p p) {
    PDU_t *pointer=&(upointer->get_request);
    int res=-1;

    if ((res=(pdu_putrefy(pointer,p)))==-1)
	err_msg("Couldn't reconstruct get-request");
    return res;
}

int
get_next_putrefy(PDU_Choice_t *upointer,mesg_data_p p) {
    PDU_t *pointer=&(upointer->get_next_request);
    int res=-1;

    if ((res=(pdu_putrefy(pointer,p)))==-1)
	err_msg("Couldn't reconstruct get-next-request");
    return res;
}



int
vbl_cr(VarBindList_t *pointer,mesg_data_p p){
    VarBind_t	*vb;
    int res=-1;
    int i;

    for (i=0;i<p->vb_n;i++) {
	vb=(VarBind_t *)Calloc(1,sizeof(VarBind_t));
	if (vb_cr(vb,p->data+i)==-1) {
	    err_msg("Couldn't construct VB");
	    return -1;
        }
        if ((res=ASN_SEQUENCE_ADD(&(pointer->list),vb))!=0){
	    err_msg("Couldn't add VB to Syntax");
	    return -1;
	}
    }
    return res;
}

int
vbl_free(VarBindList_t *pointer){
    int i;

    for (i=0;i<pointer->list.count;i++){
	if (vb_free(pointer->list.array[i])==-1){
	    err_msg("Couldn't free VB");
	    return -1;
	}
    }
    free(pointer->list.array);
    return 0;
}

int
vbl_putrefy(VarBindList_t *pointer,mesg_data_p p){
    int i;
    vb_data_p vb_data;

    vb_data=(vb_data_p)Calloc(pointer->list.count,sizeof(vb_data_s));
    p->data=vb_data;
    p->vb_n=pointer->list.count;
    for (i=0;i<pointer->list.count;i++) {
	if (vb_putrefy(pointer->list.array[i],vb_data+i)==-1){
	    err_msg("Couldn't reconstruct VB");
	    return -1;
        }
    }
    return 0;
}


int
vb_cr(VarBind_t *pointer,vb_data_p p) {
    if (oid_copy(&(pointer->name),p->oid)==-1){
	err_msg("Couldn't construct Variable Binding Name");
	return -1;
    }
    if (vbv_cr(&(pointer->value),p)==-1){
	err_msg("Couldn't construct Variable Binding Value");
	return -1;
    }
    return 0;
}

int
vb_free(VarBind_t *pointer) {
/*
    if (pointer->name.buf) {
	free(pointer->name.buf);
    }
    else {
	err_msg("No oid buffer available at Variable Binding Name");
	return -1;
    }
    if ((vbv_free(&pointer->value))==-1){
	err_msg("Couldn't free Variable Binding Value");
	return -1;
    }
*/
    asn_DEF_VarBind.free_struct(&asn_DEF_VarBind,pointer,0);
    return 0;
}

int
vb_putrefy(VarBind_t *pointer,vb_data_p p) {
    if (oid_copy(&p->oid,pointer->name)==-1){
	err_msg("Couldn't reconstruct Variable Binding Name");
	return -1;
    }
    if (vbv_putrefy(&(pointer->value),p)==-1){
	err_msg("Couldn't reconstruct Variable Binding Value");
	return -1;
    }
    return 0;
}


int
vbv_cr(VarBindValue_t *pointer,vb_data_p p) {

    pointer->present=p->vbv_present;
    if (vbv_handlers[pointer->present].vbv_cr_func) {
	if (((*vbv_handlers[pointer->present].vbv_cr_func)(&(pointer->choice),p))!=0){
	    err_msg("Couldn't construct Variable Binding Value of Choice");
	    return -1;
	}
    }
    else {
	err_msg("VBV type not implemented yet\n");
	return -1;
    }
    return 0;
}

int
vbv_free(VarBindValue_t *pointer) {
    if (vbv_handlers[pointer->present].vbv_cr_func) {
	if (((*vbv_handlers[pointer->present].vbv_free_func)(&(pointer->choice)))!=0){
	    err_msg("Couldn't free Variable Binding Value of Choice");
	    return -1;
	}
    }
    else {
	err_msg("VBV type not implemented yet\n");
	return -1;
    }
    return 0;
}

int
vbv_cpv(char * where, vb_data_p p) {
    if (vbv_handlers[p->vbv_present].vbv_cr_func) {
	if (((*vbv_handlers[p->vbv_present].vbv_cpv_func)(where,p))!=0){
	    err_msg("Couldn't copy Variable Binding Value of Choice");
	    return -1;
	}
    }
    else {
	err_msg("VBV type not implemented yet\n");
	return -1;
    }
    return 0;
}

int
vbv_putrefy(VarBindValue_t *pointer,vb_data_p p) {

    p->vbv_present=pointer->present;
    if (vbv_handlers[pointer->present].vbv_putrefy_func) {
	if (((*vbv_handlers[pointer->present].vbv_putrefy_func)(&(pointer->choice),p))!=0){
	    err_msg("Couldn't reconstruct Variable Binding Value of Choice");
	    return -1;
	}
    }
    else {
	err_msg("VBV type not implemented yet\n");
	return -1;
    }
    return 0;
}



int
vbv_unspecified_cr(VBV_Choice_t *upointer,vb_data_p p) {
    NULL_t *pointer=&(upointer->unSpecified);
    *pointer=0;
    return 0;
}

int
vbv_unspecified_free(VBV_Choice_t *upointer){
    return 0;
}

int
obj_syn_cr(VBV_Choice_t *upointer,vb_data_p p) {
    ObjectSyntax_t *pointer=&(upointer->value);

    pointer->present=p->os_present;
    if (os_handlers[pointer->present].os_cr_func) {
	if (((*os_handlers[pointer->present].os_cr_func)(&(pointer->choice),p))!=0){
	    err_msg("Couldn't construct ObjectSyntax of Choice");
	    return -1;
	}
    }
    else {
	err_msg("ObjSyn type not implemented yet\n");
	return -1;
    }
    return 0; 
}

int
obj_syn_free(VBV_Choice_t *upointer){
    ObjectSyntax_t *pointer=&(upointer->value);

    if (os_handlers[pointer->present].os_free_func) {
	if (((*os_handlers[pointer->present].os_free_func)(&(pointer->choice)))!=0){
	    err_msg("Couldn't free ObjectSyntax of Choice");
	    return -1;
	}
    }
    else {
	err_msg("ObjSyn type not implemented yet\n");
	return -1;
    }
    return 0;
}

int
obj_syn_cpv(char *where, vb_data_p p){
    if (os_handlers[p->os_present].os_cpv_func) {
	if (((*os_handlers[p->os_present].os_cpv_func)(where,p))!=0){
	    err_msg("Couldn't copy ObjectSyntax of Choice");
	    return -1;
	}
    }
    else {
	err_msg("ObjSyn type not implemented yet\n");
	return -1;
    }
    return 0; 
}

int
nothing_cpv(char *where, vb_data_p p){
    return 0;
}

int
nothing_free(VBV_Choice_t *upointer){
    return 0;
}

int
obj_syn_putrefy(VBV_Choice_t *upointer,vb_data_p p) {
    ObjectSyntax_t *pointer=&(upointer->value);

    p->os_present=pointer->present;
    if (os_handlers[pointer->present].os_putrefy_func) {
	if (((*os_handlers[pointer->present].os_putrefy_func)(&(pointer->choice),p))!=0){
	    err_msg("Couldn't reconstruct ObjectSyntax of Choice");
	    return -1;
	}
    }
    else {
	err_msg("ObjSyn type not implemented yet\n");
	return -1;
    }
    return 0; 
}

int
sim_syn_cr(OS_Choice_t *upointer,vb_data_p p) {
    SimpleSyntax_t *pointer=&(upointer->simple);

    pointer->present=p->st_present;
    
    if (ss_handlers[pointer->present].ss_cr_func) {
	if (((*ss_handlers[pointer->present].ss_cr_func)(&(pointer->choice),&(p->value)))!=0){
	    err_msg("Couldn't construct SimpleSyntax of Choice");
	    return -1;
	}
    }
    else {
	err_msg("SimSyn type not implemented yet\n");
	return -1;
    }
    return 0;
}

int
sim_syn_free(OS_Choice_t *upointer){
    SimpleSyntax_t *pointer=&(upointer->simple);

    if (ss_handlers[pointer->present].ss_free_func) {
	if (((*ss_handlers[pointer->present].ss_free_func)(&(pointer->choice)))!=0){
	    err_msg("Couldn't free SimpleSyntax of Choice");
	    return -1;
	}
    }
    else {
	err_msg("SimSyn type not implemented yet\n");
	return -1;
    }
    return 0;
}

int
sim_syn_cpv(char *where,vb_data_p p) {
    if (ss_handlers[p->st_present].ss_cr_func) {
	if (((*ss_handlers[p->st_present].ss_cpv_func)(where,p))!=0){
	    err_msg("Couldn't construct SimpleSyntax of Choice");
	    return -1;
	}
    }
    else {
	err_msg("SimSyn type not implemented yet\n");
	return -1;
    }
    return 0;
}

int
sim_syn_putrefy(OS_Choice_t *upointer,vb_data_p p) {
    SimpleSyntax_t *pointer=&(upointer->simple);

    p->st_present=pointer->present;
    if (ss_handlers[pointer->present].ss_putrefy_func) {
	if (((*ss_handlers[pointer->present].ss_putrefy_func)(&(pointer->choice),&(p->value)))!=0){
	    err_msg("Couldn't construct SimpleSyntax of Choice");
	    return -1;
	}
    }
    else {
	err_msg("SimSyn type not implemented yet\n");
	return -1;
    }
    return 0;
}


int
int_cr(SS_Choice_t *pointer,OCTET_STRING_t *data) {
    long *l=(long *)data->buf;

    pointer->integer_value=*l;
    return 0;
}

int
int_free(SS_Choice_t *pointer){
    return 0;
}

int
o_str_cr(SS_Choice_t *pointer, OCTET_STRING_t *data) {
    pointer->string_value.size=data->size;
    pointer->string_value.buf=(uint8_t *)Malloc(sizeof(uint8_t)*data->size);
    bcopy(data->buf,pointer->string_value.buf, data->size);
    return 0;
}

int
o_str_free(SS_Choice_t *pointer){
    if (!pointer->string_value.buf){
	err_msg("String without buffer to be freed");
	return -1;
    }
    free(pointer->string_value.buf);
    return 0;
}

int
oid_cr(SS_Choice_t *pointer,OCTET_STRING_t *data) {
    OBJECT_IDENTIFIER_t *oid=(OBJECT_IDENTIFIER_t *)data;

    return oid_copy(&(pointer->objectID_value),*oid);
}

int
oid_free(SS_Choice_t *pointer){
    if (!pointer->objectID_value.buf){
	err_msg("OID without buffer to be freed");
	return -1;
    }
    free(pointer->objectID_value.buf);
    return 0;
}

int
int_putrefy(SS_Choice_t *pointer,OCTET_STRING_t *data) {

    data->buf=(uint8_t *)Malloc(sizeof(long));
    data->size=sizeof(long);
    bcopy(&(pointer->integer_value),data->buf,data->size);
    return 0;
}

int
int_cpv(char * where, vb_data_p data) {

    bcopy((char*)data->value.buf,where,data->value.size);
    return 0;
}

int
o_str_putrefy(SS_Choice_t *pointer, OCTET_STRING_t *data) {

    data->buf=(uint8_t *)Malloc(pointer->string_value.size);
    data->size=pointer->string_value.size;
    bcopy(pointer->string_value.buf,data->buf,data->size);
    return 0;
}

int
o_str_cpv(char *where, vb_data_p data) {

    bcopy(data->value.buf,where,data->value.size);
    return 0;
}

int
oid_cpv(char *where, vb_data_p data) {

    bcopy(data->value.buf,where,data->value.size);
    return 0;
}


int
oid_putrefy(SS_Choice_t *pointer,OCTET_STRING_t *data ) {

    data->buf=(uint8_t *)Malloc(pointer->objectID_value.size);
    data->size=pointer->objectID_value.size;
    bcopy(pointer->objectID_value.buf,data->buf,data->size);
    return 0;
}


int
oid_create(OBJECT_IDENTIFIER_t *pointer, const char *oid) {
    long	*mib;
    int 	rb,ra;

    if ((rb=OBJECT_IDENTIFIER_parse_arcs(oid,-1,NULL,0, NULL)) <=0 ){
	err_msg("bad oid");
	return -1;
    }
    else {
	mib=(long *)Malloc(rb*arcsize);
	ra=OBJECT_IDENTIFIER_parse_arcs(oid,-1,mib,rb, NULL);
	if (ra!=rb) {
	    err_msg("couldn't parse oid");
	    free(mib);
	    return -1;
	}
	else {
	    rb = OBJECT_IDENTIFIER_set_arcs(pointer,mib,arcsize, ra);
	    if (rb==-1) {
	    	err_msg("couldn't parse oid");
	    	free(mib);
		return -1;
	    }
	    free(mib);
	}
    }
    return 0;
}

int
oid_copy(OBJECT_IDENTIFIER_t *pointer,const OBJECT_IDENTIFIER_t source) {
    
    pointer->size=source.size;
    pointer->buf=(uint8_t *)Malloc(sizeof (uint8_t)*source.size);
    bcopy(source.buf,pointer->buf,source.size);
    return 0;
}

int
os_copy(OCTET_STRING_t *pointer,OCTET_STRING_t *source) {
    
    pointer->size=source->size;
    pointer->buf=(uint8_t *)Malloc(sizeof(uint8_t)*pointer->size);
    bcopy(source->buf,pointer->buf,pointer->size);
    return 0;
}

int
notintree(OBJECT_IDENTIFIER_t *tree, OBJECT_IDENTIFIER_t *node) {
	int i;
	int min_size;
	if (tree->size > node->size)
	    min_size = node->size;
	else
	    min_size = tree->size;
	for (i=0; i<min_size; i++) {
		if (tree->buf[i] !=node->buf[i] ) return 1;
	}
	return 0;
}